/* mhs: A GObject wrapper for the Mozilla Mhs API
 *
 * Copyright (C) 2009  Intel Corporation
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <glib-object.h>
#include <dbus/dbus-glib.h>
#include <new>
#include <nsINavHistoryService.h>
#include <nsIFaviconService.h>
#include <nsINavBookmarksService.h>
#include <nsIAutoCompleteSearch.h>
#include <nsIAutoCompleteResult.h>
#include <nsIIOService.h>
#include <nsIGlobalHistory2.h>
#include <nsIBrowserHistory.h>
#include <nsToolkitCompsCID.h>
#include <nsNetCID.h>
#include <nsServiceManagerUtils.h>
#include <nsCOMPtr.h>
#include <nsStringAPI.h>
#include <nsMemory.h>
#include <nsNetUtil.h>
#include <nsDocShellCID.h>

#include "mhs-history.h"
#include "mhs-error-private.h"
#include "mhs-service.h"
#include "mhs-marshal.h"

/* D-Bus method implementations */

static gboolean mhs_history_add_uri (MhsHistory *self,
                                        const gchar *uri,
                                        gboolean redirect,
                                        gboolean top_level,
                                        const gchar *referrer,
                                        GError **error);

static gboolean mhs_history_is_visited (MhsHistory *self,
                                           const gchar *uri,
                                           gboolean *is_visited,
                                           GError **error);

static gboolean mhs_history_set_page_title (MhsHistory *self,
                                               const gchar *uri,
                                               const gchar *title,
                                               GError **error);

static gboolean mhs_history_start_ac_search (MhsHistory *self,
                                                const gchar *search_str,
                                                guint32 *search_id,
                                                GError **error);

static gboolean mhs_history_stop_ac_search (MhsHistory *self,
                                               guint32 search_id,
                                               GError **error);

static gboolean mhs_history_set_favicon_url (MhsHistory *self,
                                                const gchar *page_uri_str,
                                                const gchar *favicon_uri_str,
                                                GError **error);

static gboolean mhs_history_set_default_favicon_url
                                               (MhsHistory *self,
                                                const gchar *page_uri_str,
                                                GError **error);

static gboolean mhs_history_get_favicon (MhsHistory *self,
                                            const gchar *page_uri,
                                            gboolean download,
                                            DBusGMethodInvocation *context);

static gboolean mhs_history_get_favorites (MhsHistory *self,
                                           GError **error);

static gboolean mhs_history_remove_favorite (MhsHistory *self,
                                             const gchar *uri,
                                             GError **error);

static gboolean mhs_history_clear_history (MhsHistory *self,
                                           GError **error);

static gboolean mhs_history_get_is_page_pinned (MhsHistory *self,
                                                const gchar *page_uri,
                                                gboolean *is_page_pinned,
                                                GError **error);

static gboolean mhs_history_get_pinned_pages (MhsHistory *self,
                                              GError **error);

static gboolean mhs_history_pin_page (MhsHistory *self,
                                      const gchar *uri_str,
                                      const gchar *page_title,
                                      GError **error);

static gboolean mhs_history_unpin_page (MhsHistory *self,
                                        const gchar *uri_str,
                                        GError **error);

static gboolean mhs_history_unpin_all_pages (MhsHistory *self,
                                             GError **error);

#include "mhs-history-glue.h"

/* End D-Bus method implementations */

static void mhs_history_finalize (GObject *object);

G_DEFINE_TYPE (MhsHistory, mhs_history, G_TYPE_OBJECT);

#define MHS_HISTORY_GET_PRIVATE(obj) \
  (G_TYPE_INSTANCE_GET_PRIVATE ((obj), MHS_TYPE_HISTORY, \
                                MhsHistoryPrivate))

#define MHS_HISTORY_N_FAVORITES 9

class MhsHistoryObserver : public nsIAutoCompleteObserver,
                              public nsINavHistoryObserver
{
public:
  NS_DECL_ISUPPORTS
  NS_DECL_NSIAUTOCOMPLETEOBSERVER
  NS_DECL_NSINAVHISTORYOBSERVER

  MhsHistory *history;
};

class MhsBookmarksObserver : public nsINavBookmarkObserver
{
public:
  NS_DECL_ISUPPORTS
  NS_DECL_NSINAVBOOKMARKOBSERVER

  MhsHistory *history;
};

struct _MhsHistoryPrivate
{
  guint32 ac_search_id;
  gboolean in_ac_search;

  /* Number of history items that we have reported back from the
     result set so far */
  guint n_results_reported;

  gboolean history_observer_added;
  MhsHistoryObserver observer;

  gboolean bookmarks_observer_added;
  MhsBookmarksObserver bookmarks_observer;

  GSList *favicon_closures;

  /* Idle handler for fetching the favorites */
  guint favorites_idle_handler;
  /* and for fetching the pinned pages */
  guint pinned_pages_idle_handler;
};

struct MhsHistoryFaviconClosure
{
  MhsHistoryFaviconClosure (nsIURI *_page_uri, nsIURI *_favicon_uri,
                               DBusGMethodInvocation *_context)
    : page_uri (_page_uri),
      favicon_uri (_favicon_uri),
      context (_context)
  {
  }

  nsCOMPtr<nsIURI> page_uri, favicon_uri;
  DBusGMethodInvocation *context;
};

enum
  {
    AC_RESULT_RECEIVED_SIGNAL,
    LINK_VISITED_SIGNAL,
    FAVORITES_RECEIVED_SIGNAL,
    PINNED_PAGE_SIGNAL,
    UNPINNED_PAGE_SIGNAL,

    LAST_SIGNAL
  };

static guint history_signals[LAST_SIGNAL] = { 0, };

NS_IMPL_QUERY_INTERFACE2 (MhsHistoryObserver,
                          nsIAutoCompleteObserver,
                          nsINavHistoryObserver)

/* We don't want to implement reference counting for
   MhsHistoryObserver because it is entirely owned by the
   MhsHistory instance */
NS_IMETHODIMP_(nsrefcnt)
MhsHistoryObserver::AddRef ()
{
  return 1;
}

NS_IMETHODIMP_(nsrefcnt)
MhsHistoryObserver::Release ()
{
  return 1;
}

static void
mhs_history_favicon_closure_destroy (MhsHistoryFaviconClosure *data)
{
  if (data->context)
    {
      GError *error = NULL;
      mhs_error_set_from_nsresult (NS_ERROR_NOT_AVAILABLE, &error);
      dbus_g_method_return_error (data->context, error);
      g_error_free (error);
    }

  data->~MhsHistoryFaviconClosure ();
  g_slice_free (MhsHistoryFaviconClosure, data);
}

NS_IMETHODIMP
MhsHistoryObserver::OnSearchResult (nsIAutoCompleteSearch *search,
                                       nsIAutoCompleteResult *result)
{
  nsresult rv;
  MhsHistoryPrivate *priv = history->priv;
  PRUint16 search_result;
  PRUint32 match_count;

  rv = result->GetSearchResult (&search_result);

  if (NS_FAILED (rv))
    {
      g_warning ("GetSearchResult failed: 0x%x\n", rv);
      return rv;
    }

  rv = result->GetMatchCount (&match_count);

  if (NS_FAILED (rv))
    {
      g_warning ("GetMatchCount failed: 0x%x\n", rv);
      return rv;
    }

  if (search_result == nsIAutoCompleteResult::RESULT_SUCCESS
      || search_result == nsIAutoCompleteResult::RESULT_SUCCESS_ONGOING)
    while (priv->n_results_reported < match_count)
      {
        nsAutoString value, comment, image;

        if (NS_SUCCEEDED (result->GetValueAt (priv->n_results_reported, value))
            && NS_SUCCEEDED (result->GetCommentAt (priv->n_results_reported,
                                                   comment)))
          {
            NS_ConvertUTF16toUTF8 value_utf8 (value);
            NS_ConvertUTF16toUTF8 comment_utf8 (comment);

            g_signal_emit (history,
                           history_signals[AC_RESULT_RECEIVED_SIGNAL], 0,
                           priv->ac_search_id,
                           value_utf8.get (),
                           comment_utf8.get ());
          }

        priv->n_results_reported++;
      }

  return NS_OK;
}

NS_IMETHODIMP
MhsHistoryObserver::OnBeginUpdateBatch ()
{
  return NS_ERROR_NOT_IMPLEMENTED;
}

NS_IMETHODIMP
MhsHistoryObserver::OnEndUpdateBatch ()
{
  return NS_ERROR_NOT_IMPLEMENTED;
}

NS_IMETHODIMP
MhsHistoryObserver::OnVisit (nsIURI *uri,
                                PRInt64 visit_id,
                                PRTime when,
                                PRInt64 session_id,
                                PRInt64 referring_id,
                                PRUint32 transition_type,
                                PRUint32 *added NS_OUTPARAM)
{
  gint visit_time = when / G_USEC_PER_SEC;
  nsCAutoString uri_spec;
  nsresult rv;

  rv = uri->GetSpec (uri_spec);
  if (NS_FAILED (rv))
    return rv;

  // Signal that a link has been visited
  g_signal_emit (history, history_signals[LINK_VISITED_SIGNAL], 0,
                 uri_spec.get (), visit_time);

  return NS_OK;
}

NS_IMETHODIMP
MhsHistoryObserver::OnTitleChanged (nsIURI *uri,
                                       const nsAString &page_title)
{
  return NS_ERROR_NOT_IMPLEMENTED;
}

NS_IMETHODIMP
MhsHistoryObserver::OnBeforeDeleteURI(nsIURI* uri)
{
  return NS_ERROR_NOT_IMPLEMENTED;
}

NS_IMETHODIMP
MhsHistoryObserver::OnDeleteURI (nsIURI *uri)
{
  return NS_ERROR_NOT_IMPLEMENTED;
}

NS_IMETHODIMP
MhsHistoryObserver::OnClearHistory ()
{
  return NS_ERROR_NOT_IMPLEMENTED;
}

NS_IMETHODIMP
MhsHistoryObserver::OnPageChanged (nsIURI *uri,
                                      PRUint32 what_changed,
                                      const nsAString &value)
{
  MhsHistoryPrivate *priv = history->priv;
  nsresult rv;

  if (what_changed == ATTRIBUTE_FAVICON)
    {
      GSList *next;

      nsCOMPtr<nsIFaviconService> favicon_service
        = do_GetService (NS_FAVICONSERVICE_CONTRACTID, &rv);

      if (NS_FAILED (rv))
        return rv;

      /* Look for a matching URI in the list of closures we are
         waiting for */
      for (GSList *l = priv->favicon_closures; l; l = next)
        {
          MhsHistoryFaviconClosure *data
            = (MhsHistoryFaviconClosure *) l->data;
          PRBool uri_equals;

          next = l->next;

          rv = data->page_uri->Equals (uri, &uri_equals);

          if (NS_SUCCEEDED (rv) && uri_equals)
            {
              nsCAutoString mime_type;
              PRUint32 data_len;
              PRUint8 *icon_data;

              rv = favicon_service->GetFaviconData (data->favicon_uri,
                                                    mime_type,
                                                    &data_len, &icon_data);

              if (NS_FAILED (rv))
                {
                  GError *error = NULL;
                  mhs_error_set_from_nsresult (rv, &error);
                  dbus_g_method_return_error (data->context, error);
                  g_error_free (error);
                }
              else
                {
                  /* Copy the data into a GArray to pass to D-BUS */
                  GArray *data_array =
                    g_array_sized_new (FALSE, FALSE, sizeof (guint8), data_len);

                  g_array_append_vals (data_array, icon_data, data_len);

                  dbus_g_method_return (data->context,
                                        mime_type.get (),
                                        data_array);

                  g_array_free (data_array, TRUE);
                  nsMemory::Free (icon_data);
                }

              data->context = NULL;

              priv->favicon_closures
                = g_slist_remove (priv->favicon_closures, data);
              mhs_history_favicon_closure_destroy (data);
            }
        }
    }

  return NS_OK;
}

NS_IMETHODIMP
MhsHistoryObserver::OnPageExpired (nsIURI *uri,
                                      PRTime visit_time,
                                      PRBool whole_entry)
{
  return NS_ERROR_NOT_IMPLEMENTED;
}

NS_IMPL_QUERY_INTERFACE1 (MhsBookmarksObserver,
                          nsINavBookmarkObserver)

/* We don't want to implement reference counting for
   MhsBookmarksObserver because it is entirely owned by the
   MhsHistory instance */
NS_IMETHODIMP_(nsrefcnt)
MhsBookmarksObserver::AddRef ()
{
  return 1;
}

NS_IMETHODIMP_(nsrefcnt)
MhsBookmarksObserver::Release ()
{
  return 1;
}

NS_IMETHODIMP
MhsBookmarksObserver::OnBeginUpdateBatch ()
{
  return NS_ERROR_NOT_IMPLEMENTED;
}

NS_IMETHODIMP
MhsBookmarksObserver::OnEndUpdateBatch ()
{
  return NS_ERROR_NOT_IMPLEMENTED;
}

static nsresult
mhs_history_emit_pinned_page_for_item (MhsHistory *self,
                                       nsCOMPtr<nsINavBookmarksService>
                                       bookmark_service,
                                       nsCOMPtr<nsINavHistoryService>
                                       history_service,
                                       PRInt64 bookmark_id,
                                       gboolean more_pending)
{
  nsresult rv = NS_OK;
  nsCAutoString title, uri_str;
  PRUint16 item_type;
  nsCOMPtr<nsIURI> uri;

  rv = bookmark_service->GetItemType (bookmark_id, &item_type);
  if (NS_FAILED (rv))
    goto rv_error;

  if (item_type == nsINavBookmarksService::TYPE_BOOKMARK)
    {
      nsCOMPtr<nsINavHistoryQuery> query;
      nsCOMPtr<nsINavHistoryQueryOptions> query_options;
      nsCOMPtr<nsINavHistoryResult> query_results;
      nsCOMPtr<nsINavHistoryContainerResultNode> root_result;
      nsCOMPtr<nsINavHistoryResultNode> result_node;
      PRTime visit_time = 0;

      rv = bookmark_service->GetItemTitle (bookmark_id, title);
      if (NS_FAILED (rv))
        goto rv_error;
      rv = bookmark_service->GetBookmarkURI (bookmark_id,
                                             getter_AddRefs (uri));
      if (NS_FAILED (rv))
        goto rv_error;

      /* Query for the last visit time to this URI */
      rv = history_service->GetNewQuery (getter_AddRefs (query));
      if (NS_FAILED (rv))
        goto rv_error;
      rv = history_service->GetNewQueryOptions (getter_AddRefs (query_options));
      if (NS_FAILED (rv))
        goto rv_error;

      rv = query->SetUri (uri);
      if (NS_FAILED (rv))
        goto rv_error;
      rv = query_options->SetResultType (nsINavHistoryQueryOptions
                                         ::RESULTS_AS_URI);
      if (NS_FAILED (rv))
        goto rv_error;

      rv = history_service->ExecuteQuery (query, query_options,
                                          getter_AddRefs (query_results));
      if (NS_FAILED (rv))
        goto rv_error;

      rv = query_results->GetRoot (getter_AddRefs (root_result));
      if (NS_FAILED (rv))
        goto rv_error;

      rv = root_result->SetContainerOpen (PR_TRUE);
      if (NS_FAILED (rv))
        goto rv_error;

      rv = root_result->GetChild (0, getter_AddRefs (result_node));
      if (NS_SUCCEEDED (rv))
        {
          rv = result_node->GetTime (&visit_time);
          if (NS_FAILED (rv))
            goto rv_error;
        }
      /* If there was no entry then it's not an error and we can just
         leave the visit time as zero */
      else if (rv != NS_ERROR_INVALID_ARG)
        goto rv_error;

      rv = uri->GetSpec (uri_str);
      if (NS_FAILED (rv))
        goto rv_error;

      g_signal_emit (self, history_signals[PINNED_PAGE_SIGNAL], 0,
                     title.get (), uri_str.get (),
                     (gint) (visit_time / G_USEC_PER_SEC),
                     more_pending);
    }

 rv_error:
  return rv;
}

NS_IMETHODIMP
MhsBookmarksObserver::OnItemAdded (PRInt64 item_id,
                                   PRInt64 folder,
                                   PRInt32 index)
{
  nsresult rv;
  nsCOMPtr<nsINavBookmarksService> bookmark_service;
  nsCOMPtr<nsINavHistoryService> history_service;

  bookmark_service = do_GetService (NS_NAVBOOKMARKSSERVICE_CONTRACTID, &rv);
  if (NS_FAILED (rv))
    return rv;
  history_service = do_GetService (NS_NAVHISTORYSERVICE_CONTRACTID, &rv);
  if (NS_FAILED (rv))
    return rv;

  mhs_history_emit_pinned_page_for_item (history,
                                         bookmark_service,
                                         history_service,
                                         item_id,
                                         FALSE);
  if (NS_FAILED (rv))
    return rv;

  return NS_OK;
}

NS_IMETHODIMP
MhsBookmarksObserver::OnBeforeItemRemoved (PRInt64 item_id)
{
  nsCOMPtr<nsINavBookmarksService> bookmark_service;
  nsresult rv;
  nsCAutoString uri_str;
  PRUint16 item_type;
  nsCOMPtr<nsIURI> uri;

  bookmark_service = do_GetService (NS_NAVBOOKMARKSSERVICE_CONTRACTID, &rv);
  if (NS_FAILED (rv))
    return rv;

  rv = bookmark_service->GetItemType (item_id, &item_type);
  if (NS_FAILED (rv))
    return rv;

  if (item_type == nsINavBookmarksService::TYPE_BOOKMARK)
    {
      rv = bookmark_service->GetBookmarkURI (item_id,
                                             getter_AddRefs (uri));
      if (NS_FAILED (rv))
        return rv;
      rv = uri->GetSpec (uri_str);
      if (NS_FAILED (rv))
        return rv;

      g_signal_emit (history, history_signals[UNPINNED_PAGE_SIGNAL], 0,
                     uri_str.get ());
    }

  return NS_OK;
}

NS_IMETHODIMP
MhsBookmarksObserver::OnItemRemoved (PRInt64 item_id,
                                     PRInt64 folder,
                                     PRInt32 index)
{
  return NS_ERROR_NOT_IMPLEMENTED;
}

NS_IMETHODIMP
MhsBookmarksObserver::OnItemChanged (PRInt64 bookmark_id,
                                     const nsACString &property,
                                     PRBool is_annotation_property,
                                     const nsACString &value)
{
  return NS_ERROR_NOT_IMPLEMENTED;
}

NS_IMETHODIMP
MhsBookmarksObserver::OnItemVisited (PRInt64 bookmark_id,
                                     PRInt64 visit_id,
                                     PRTime time)
{
  return NS_ERROR_NOT_IMPLEMENTED;
}

NS_IMETHODIMP
MhsBookmarksObserver::OnItemMoved (PRInt64 item_id,
                                   PRInt64 old_parent,
                                   PRInt32 old_index,
                                   PRInt64 new_parent,
                                   PRInt32 new_index)
{
  return NS_ERROR_NOT_IMPLEMENTED;
}

static void
mhs_history_class_init (MhsHistoryClass *klass)
{
  GObjectClass *gobject_class = (GObjectClass *) klass;

  gobject_class->finalize = mhs_history_finalize;

  history_signals[AC_RESULT_RECEIVED_SIGNAL] =
    g_signal_new ("ac-result-received",
                  G_TYPE_FROM_CLASS (gobject_class),
                  G_SIGNAL_RUN_LAST,
                  G_STRUCT_OFFSET (MhsHistoryClass, ac_result_received),
                  NULL, NULL,
                  mhs_marshal_VOID__UINT_STRING_STRING,
                  G_TYPE_NONE, 3,
                  G_TYPE_UINT,
                  G_TYPE_STRING,
                  G_TYPE_STRING);

  history_signals[LINK_VISITED_SIGNAL] =
    g_signal_new ("link-visited",
                  G_TYPE_FROM_CLASS (gobject_class),
                  G_SIGNAL_RUN_LAST,
                  G_STRUCT_OFFSET (MhsHistoryClass, link_visited),
                  NULL, NULL,
                  mhs_marshal_VOID__STRING_INT,
                  G_TYPE_NONE, 2, G_TYPE_STRING, G_TYPE_INT);

  history_signals[FAVORITES_RECEIVED_SIGNAL] =
    g_signal_new ("favorites-received",
                  G_TYPE_FROM_CLASS (gobject_class),
                  G_SIGNAL_RUN_LAST,
                  G_STRUCT_OFFSET (MhsHistoryClass, favorites_received),
                  NULL, NULL,
                  mhs_marshal_VOID__BOXED_BOXED,
                  G_TYPE_NONE, 2, G_TYPE_STRV, G_TYPE_STRV);

  history_signals[PINNED_PAGE_SIGNAL] =
    g_signal_new ("pinned-page",
                  G_TYPE_FROM_CLASS (gobject_class),
                  G_SIGNAL_RUN_LAST,
                  G_STRUCT_OFFSET (MhsHistoryClass, pinned_page),
                  NULL, NULL,
                  mhs_marshal_VOID__STRING_STRING_INT_BOOL,
                  G_TYPE_NONE, 4,
                  G_TYPE_STRING, G_TYPE_STRING, G_TYPE_INT, G_TYPE_BOOLEAN);

  history_signals[UNPINNED_PAGE_SIGNAL] =
    g_signal_new ("unpinned-page",
                  G_TYPE_FROM_CLASS (gobject_class),
                  G_SIGNAL_RUN_LAST,
                  G_STRUCT_OFFSET (MhsHistoryClass, unpinned_page),
                  NULL, NULL,
                  g_cclosure_marshal_VOID__STRING,
                  G_TYPE_NONE, 1, G_TYPE_STRING);

  g_type_class_add_private (klass, sizeof (MhsHistoryPrivate));

  dbus_g_object_type_install_info (MHS_TYPE_HISTORY,
                                   &dbus_glib_mhs_history_object_info);
}

static void
mhs_history_init (MhsHistory *self)
{
  MhsHistoryPrivate *priv;
  DBusGConnection *connection;
  nsresult rv;
  GError *error = NULL;

  priv = self->priv = MHS_HISTORY_GET_PRIVATE (self);

  priv->ac_search_id = 1;

  /* 'placement new' to call the constructor for
     MhsHistoryPrivate */
  new (reinterpret_cast<void *> (priv)) MhsHistoryPrivate;

  if ((connection = dbus_g_bus_get (DBUS_BUS_SESSION, &error)) == NULL)
    {
      g_warning ("Error connecting to session bus: %s", error->message);
      g_error_free (error);
    }
  else
    {
      dbus_g_connection_register_g_object (connection,
                                           MHS_SERVICE_HISTORY_PATH,
                                           G_OBJECT (self));
      dbus_g_connection_unref (connection);
    }

  priv->observer.history = self;
  priv->bookmarks_observer.history = self;

  /* Register the observer for bookmark notifications */
  nsCOMPtr<nsINavBookmarksService> bookmarks_service
    = do_GetService (NS_NAVBOOKMARKSSERVICE_CONTRACTID, &rv);
  if (NS_SUCCEEDED (rv))
    {
      rv = bookmarks_service->AddObserver (&priv->bookmarks_observer,
                                           PR_FALSE);

      if (NS_SUCCEEDED (rv))
        priv->bookmarks_observer_added = TRUE;
    }

  /* Register for history notifications */
  nsCOMPtr<nsINavHistoryService> history_service
    = do_GetService (NS_NAVHISTORYSERVICE_CONTRACTID, &rv);
  if (NS_SUCCEEDED (rv))
    {
      rv = history_service->AddObserver (&priv->observer, PR_FALSE);

      if (NS_SUCCEEDED (rv))
        priv->history_observer_added = TRUE;
    }
}

static void
mhs_history_finalize (GObject *object)
{
  MhsHistory *self = (MhsHistory *) object;
  MhsHistoryPrivate *priv = self->priv;
  nsresult rv;

  mhs_history_stop_ac_search (self, priv->ac_search_id, NULL);

  if (priv->history_observer_added)
    {
      /* Unregister the observer for history notifications as the observer
         object is going to be destroyed */
      nsCOMPtr<nsINavHistoryService> history_service
        = do_GetService (NS_NAVHISTORYSERVICE_CONTRACTID, &rv);
      if (NS_SUCCEEDED (rv))
        history_service->RemoveObserver (&priv->observer);
    }
  if (priv->bookmarks_observer_added)
    {
      /* Same for bookmark notifications */
      nsCOMPtr<nsINavBookmarksService> bookmarks_service
        = do_GetService (NS_NAVBOOKMARKSSERVICE_CONTRACTID, &rv);
      if (NS_SUCCEEDED (rv))
        bookmarks_service->RemoveObserver (&priv->bookmarks_observer);
    }

  g_slist_foreach (priv->favicon_closures,
                   (GFunc) mhs_history_favicon_closure_destroy, NULL);
  g_slist_free (priv->favicon_closures);
  priv->favicon_closures = NULL;

  if (priv->favorites_idle_handler)
    {
      g_source_remove (priv->favorites_idle_handler);
      priv->favorites_idle_handler = 0;
    }

  if (priv->pinned_pages_idle_handler)
    {
      g_source_remove (priv->pinned_pages_idle_handler);
      priv->pinned_pages_idle_handler = 0;
    }

  /* Explicitly call the destructor for the private data (so that it
     destructs without trying to free the memory) */
  priv->~MhsHistoryPrivate ();

  G_OBJECT_CLASS (mhs_history_parent_class)->finalize (object);
}

MhsHistory *
mhs_history_new (void)
{
  MhsHistory *self = (MhsHistory *) g_object_new (MHS_TYPE_HISTORY,
                                                        NULL);

  return self;
}

static gboolean
mhs_history_add_uri (MhsHistory *self,
                        const gchar *uri,
                        gboolean redirect,
                        gboolean top_level,
                        const gchar *referrer,
                        GError **error)
{
  nsresult rv;

  g_return_val_if_fail (MHS_IS_HISTORY (self), FALSE);

  nsCOMPtr<nsIGlobalHistory2> global_history2
    = do_GetService (NS_NAVHISTORYSERVICE_CONTRACTID, &rv);

  if (NS_FAILED (rv))
    {
      mhs_error_set_from_nsresult (rv, error);
      return FALSE;
    }

  nsCOMPtr<nsIURI> nsuri;
  rv = NS_NewURI (getter_AddRefs (nsuri), uri);
  if (NS_FAILED (rv))
    {
      mhs_error_set_from_nsresult (rv, error);
      return FALSE;
    }

  nsCOMPtr<nsIURI> referrer_nsuri;
  if (redirect)
    {
      rv = NS_NewURI (getter_AddRefs (referrer_nsuri), uri);
      if (NS_FAILED (rv))
        {
          mhs_error_set_from_nsresult (rv, error);
          return FALSE;
        }
    }

  rv = global_history2->AddURI (nsuri, redirect, top_level,
                                referrer_nsuri);

  if (NS_FAILED (rv))
    {
      mhs_error_set_from_nsresult (rv, error);
      return FALSE;
    }

  return TRUE;
}

static gboolean
mhs_history_is_visited (MhsHistory *self,
                           const gchar *uri,
                           gboolean *is_visited,
                           GError **error)
{
  nsresult rv;

  g_return_val_if_fail (MHS_IS_HISTORY (self), FALSE);

  nsCOMPtr<nsIGlobalHistory2> global_history2
    = do_GetService (NS_NAVHISTORYSERVICE_CONTRACTID, &rv);

  if (NS_FAILED (rv))
    {
      mhs_error_set_from_nsresult (rv, error);
      return FALSE;
    }

  nsCOMPtr<nsIURI> nsuri;
  rv = NS_NewURI (getter_AddRefs (nsuri), uri);
  if (NS_FAILED (rv))
    {
      mhs_error_set_from_nsresult (rv, error);
      return FALSE;
    }

  if (is_visited)
    {
      PRBool result;
      rv = global_history2->IsVisited (nsuri, &result);

      if (NS_FAILED (rv))
        {
          mhs_error_set_from_nsresult (rv, error);
          return FALSE;
        }

      *is_visited = result;
    }

  return TRUE;
}

static gboolean
mhs_history_set_page_title (MhsHistory *self,
                               const gchar *uri,
                               const gchar *title,
                               GError **error)
{
  nsresult rv;

  nsCOMPtr<nsIGlobalHistory2> global_history2
    = do_GetService (NS_NAVHISTORYSERVICE_CONTRACTID, &rv);

  if (NS_FAILED (rv))
    {
      mhs_error_set_from_nsresult (rv, error);
      return FALSE;
    }

  nsCOMPtr<nsIURI> nsuri;
  rv = NS_NewURI (getter_AddRefs (nsuri), uri);
  if (NS_FAILED (rv))
    {
      mhs_error_set_from_nsresult (rv, error);
      return FALSE;
    }

  nsDependentCString title_cstring (title);
  NS_ConvertUTF8toUTF16 title_string (title_cstring);

  /* The redirect parameter is deprecated so we always set it to false */
  rv = global_history2->SetPageTitle (nsuri, title_string);

  if (NS_FAILED (rv))
    {
      mhs_error_set_from_nsresult (rv, error);
      return FALSE;
    }

  return TRUE;
}

static gboolean
mhs_history_start_ac_search (MhsHistory *self,
                                const gchar *search_str,
                                guint32 *search_id,
                                GError **error)
{
  MhsHistoryPrivate *priv = self->priv;
  nsresult rv;

  nsCOMPtr<nsIAutoCompleteSearch> auto_complete
    = do_GetService (NS_NAVHISTORYSERVICE_CONTRACTID, &rv);

  if (NS_FAILED (rv))
    {
      mhs_error_set_from_nsresult (rv, error);
      return FALSE;
    }

  /* Stop any existing searches */
  rv = auto_complete->StopSearch ();
  if (NS_FAILED (rv))
    {
      mhs_error_set_from_nsresult (rv, error);
      return FALSE;
    }

  priv->n_results_reported = 0;

  rv = auto_complete->StartSearch (NS_ConvertUTF8toUTF16 (search_str),
                                   EmptyString (),
                                   NULL,
                                   &priv->observer);
  if (NS_FAILED (rv))
    {
      mhs_error_set_from_nsresult (rv, error);
      return FALSE;
    }

  priv->in_ac_search = TRUE;

  *search_id = priv->ac_search_id;

  return TRUE;
}

static gboolean
mhs_history_stop_ac_search (MhsHistory *self,
                               guint32 search_id,
                               GError **error)
{
  nsresult rv;

  g_return_val_if_fail (MHS_IS_HISTORY (self), FALSE);

  MhsHistoryPrivate *priv = self->priv;

  if (priv->in_ac_search && priv->ac_search_id == search_id)
    {
      nsCOMPtr<nsIAutoCompleteSearch> auto_complete
        = do_GetService (NS_NAVHISTORYSERVICE_CONTRACTID, &rv);

      if (NS_FAILED (rv))
        {
          mhs_error_set_from_nsresult (rv, error);
          return FALSE;
        }

      rv = auto_complete->StopSearch ();
      if (NS_FAILED (rv))
        {
          mhs_error_set_from_nsresult (rv, error);
          return FALSE;
        }

      priv->in_ac_search = FALSE;

      /* Increase the search id so that we know to ignore any results
         for previous queries. In the unlikely event of wrap-around,
         we don't want to reuse search id 0 */
      if (++priv->ac_search_id == 0)
        priv->ac_search_id = 1;
    }

  return TRUE;
}

static gboolean
mhs_history_set_favicon_url (MhsHistory *self,
                                const gchar *page_uri_str,
                                const gchar *favicon_uri_str,
                                GError **error)
{
  nsresult rv;

  g_return_val_if_fail (MHS_IS_HISTORY (self), FALSE);

  nsCOMPtr<nsIURI> favicon_uri, page_uri;

  nsCOMPtr<nsIFaviconService> favicon_service
    = do_GetService (NS_FAVICONSERVICE_CONTRACTID, &rv);

  if (NS_FAILED (rv))
    {
      mhs_error_set_from_nsresult (rv, error);
      return FALSE;
    }

  rv = NS_NewURI (getter_AddRefs (page_uri), page_uri_str);
  if (NS_FAILED (rv))
    {
      mhs_error_set_from_nsresult (rv, error);
      return FALSE;
    }
  rv = NS_NewURI (getter_AddRefs (favicon_uri), favicon_uri_str);
  if (NS_FAILED (rv))
    {
      mhs_error_set_from_nsresult (rv, error);
      return FALSE;
    }

  rv = favicon_service->SetFaviconUrlForPage (page_uri, favicon_uri);

  if (NS_FAILED (rv))
    {
      mhs_error_set_from_nsresult (rv, error);
      return FALSE;
    }

  return TRUE;
}

static gboolean
mhs_history_set_default_favicon_url (MhsHistory *self,
                                        const gchar *page_uri_str,
                                        GError **error)
{
  nsresult rv;

  g_return_val_if_fail (MHS_IS_HISTORY (self), FALSE);

  nsCOMPtr<nsIURI> favicon_uri, page_uri;
  nsCAutoString uri_path;

  nsCOMPtr<nsIFaviconService> favicon_service
    = do_GetService (NS_FAVICONSERVICE_CONTRACTID, &rv);

  if (NS_FAILED (rv))
    {
      mhs_error_set_from_nsresult (rv, error);
      return FALSE;
    }

  rv = NS_NewURI (getter_AddRefs (page_uri), page_uri_str);
  if (NS_FAILED (rv))
    {
      mhs_error_set_from_nsresult (rv, error);
      return FALSE;
    }

  /* Make a new URL with the pre_path of the page URI + "/favicon.uri" */
  rv = page_uri->GetPrePath (uri_path);
  if (NS_FAILED (rv))
    {
      mhs_error_set_from_nsresult (rv, error);
      return FALSE;
    }

  uri_path.AppendLiteral ("/favicon.ico");

  rv = NS_NewURI (getter_AddRefs (favicon_uri), uri_path);
  if (NS_FAILED (rv))
    {
      mhs_error_set_from_nsresult (rv, error);
      return FALSE;
    }

  rv = favicon_service->SetFaviconUrlForPage (page_uri, favicon_uri);

  if (NS_FAILED (rv))
    {
      mhs_error_set_from_nsresult (rv, error);
      return FALSE;
    }

  return TRUE;
}

static gboolean
mhs_history_get_favicon (MhsHistory *self,
                            const gchar *page_uri_str,
                            gboolean download,
                            DBusGMethodInvocation *context)
{
  nsresult rv;
  GError *error = NULL;

  g_return_val_if_fail (MHS_IS_HISTORY (self), FALSE);

  MhsHistoryPrivate *priv = self->priv;
  nsCOMPtr<nsIURI> favicon_uri, page_uri;

  nsCOMPtr<nsIFaviconService> favicon_service
    = do_GetService (NS_FAVICONSERVICE_CONTRACTID, &rv);

  if (NS_FAILED (rv))
    goto rv_error;

  rv = NS_NewURI (getter_AddRefs (page_uri), page_uri_str);
  if (NS_FAILED (rv))
    goto rv_error;

  /* Get the URL for the favicon of this page or bail out if it hasn't
     been set yet */
  rv = favicon_service->GetFaviconForPage (page_uri,
                                           getter_AddRefs (favicon_uri));
  if (NS_FAILED (rv))
    goto rv_error;

  /* Check if mhs already has the data for the icon cached */
  {
    nsCAutoString mime_type;
    PRUint32 data_len;
    PRUint8 *data = nsnull;

    rv = favicon_service->GetFaviconData (favicon_uri, mime_type,
                                          &data_len, &data);

    /* The documentation says that GetFaviconData will
       return NS_ERROR_NOT_AVAILABLE is there is no data, but it
       actually appears to return an empty array instead */
    if (NS_SUCCEEDED (rv) && data_len == 0)
      {
        if (data != nsnull)
          nsMemory::Free (data);
        rv = NS_ERROR_NOT_AVAILABLE;
      }

    /* If it isn't available then start downloading it
       asynchronously */
    if (rv == NS_ERROR_NOT_AVAILABLE && download)
      {
        MhsHistoryFaviconClosure *closure
          = g_slice_new (MhsHistoryFaviconClosure);
        new (reinterpret_cast<void *> (closure))
          MhsHistoryFaviconClosure (page_uri, favicon_uri, context);
        priv->favicon_closures = g_slist_prepend (priv->favicon_closures, closure);

        rv = favicon_service->SetAndLoadFaviconForPage (page_uri, favicon_uri,
                                                        PR_FALSE);
        if (NS_FAILED (rv))
          {
            priv->favicon_closures
              = g_slist_remove (priv->favicon_closures, closure);
            closure->context = NULL;
            mhs_history_favicon_closure_destroy (closure);
            goto rv_error;
          }
      }
    else if (NS_FAILED (rv))
      goto rv_error;
    else
      {
        /* Copy the data into a GArray to pass to D-BUS */
        GArray *data_array =
          g_array_sized_new (FALSE, FALSE, sizeof (guint8), data_len);

        g_array_append_vals (data_array, data, data_len);

        dbus_g_method_return (context,
                              mime_type.get (),
                              data_array);

        g_array_free (data_array, TRUE);
        nsMemory::Free (data);
      }
  }

  return TRUE;

 rv_error:
  mhs_error_set_from_nsresult (rv, &error);
  dbus_g_method_return_error (context, error);
  g_error_free (error);
  return TRUE;
}

static gboolean
mhs_history_favorites_cb (gpointer data)
{
  nsresult rv = NS_OK;
  MhsHistory *self = (MhsHistory *) data;
  MhsHistoryPrivate *priv = self->priv;
  gchar **urls, **titles;
  PRUint32 root_child_count;

  priv->favorites_idle_handler = 0;

  nsCOMPtr<nsINavHistoryService> history_service;
  nsCOMPtr<nsINavHistoryQuery> query;
  nsCOMPtr<nsINavHistoryQueryOptions> query_options;
  nsCOMPtr<nsINavHistoryResult> result;
  nsCOMPtr<nsINavHistoryContainerResultNode> root_node;

  history_service = do_GetService (NS_NAVHISTORYSERVICE_CONTRACTID, &rv);
  if (NS_FAILED (rv))
    goto rv_error;

  rv = history_service->GetNewQuery (getter_AddRefs (query));
  if (NS_FAILED (rv))
    goto rv_error;

  rv = history_service->GetNewQueryOptions (getter_AddRefs (query_options));
  if (NS_FAILED (rv))
    goto rv_error;

  query_options->SetSortingMode (nsINavHistoryQueryOptions
                                 ::SORT_BY_VISITCOUNT_DESCENDING);
  if (NS_FAILED (rv))
    goto rv_error;
  query_options->SetResultType (nsINavHistoryQueryOptions
                                ::RESULTS_AS_SITE_QUERY);
  if (NS_FAILED (rv))
    goto rv_error;
  /* Times two here in case we encounter errors while iterating over
   * the result (i.e. fallback results)
   */
  query_options->SetMaxResults (MHS_HISTORY_N_FAVORITES * 2);
  if (NS_FAILED (rv))
    goto rv_error;
  query_options->SetQueryType (nsINavHistoryQueryOptions
                               ::QUERY_TYPE_HISTORY);
  if (NS_FAILED (rv))
    goto rv_error;
  query_options->SetExpandQueries (PR_TRUE);
  if (NS_FAILED (rv))
    goto rv_error;

  rv = history_service->ExecuteQuery (query, query_options,
                                      getter_AddRefs (result));
  if (NS_FAILED (rv))
    goto rv_error;

  rv = result->GetRoot (getter_AddRefs (root_node));
  if (NS_FAILED (rv))
    goto rv_error;

  rv = root_node->SetContainerOpen (PR_TRUE);
  if (NS_FAILED (rv))
    goto rv_error;

  urls = (gchar **) g_malloc0 (sizeof (gchar *)
                               * (MHS_HISTORY_N_FAVORITES + 1));
  titles = (gchar **) g_malloc0 (sizeof (gchar *)
                                 * (MHS_HISTORY_N_FAVORITES + 1));

  rv = root_node->GetChildCount (&root_child_count);
  if (NS_SUCCEEDED (rv))
    {
      PRUint32 outer, n_found;

      for (outer = 0, n_found = 0;
           (outer < root_child_count) && (n_found < MHS_HISTORY_N_FAVORITES);
           outer++)
        {
          nsCOMPtr<nsINavHistoryResultNode> node, best_node;
          nsCOMPtr<nsINavHistoryContainerResultNode> container_node;
          PRUint32 best_access_count = 0, inner, child_count;
          nsCAutoString url, title;

          rv = root_node->GetChild (outer, getter_AddRefs (node));
          if (NS_FAILED (rv))
            break;
          container_node = do_QueryInterface (node, &rv);
          if (NS_FAILED (rv))
            break;

          /* Find the child of this node that has the highest visit
             count */
          rv = container_node->SetContainerOpen (PR_TRUE);
          if (NS_FAILED (rv))
            break;
          rv = container_node->GetChildCount (&child_count);
          if (NS_FAILED (rv))
            break;

          for (inner = 0; inner < child_count; inner++)
            {
              nsCOMPtr<nsINavHistoryResultNode> inner_node;
              PRUint32 inner_access_count;

              rv = container_node->GetChild (inner,
                                             getter_AddRefs (inner_node));
              if (NS_FAILED (rv))
                break;

              rv = inner_node->GetAccessCount (&inner_access_count);
              if (NS_FAILED (rv))
                break;

              if (inner_access_count > best_access_count)
                {
                  best_access_count = inner_access_count;
                  best_node = inner_node;
                }
            }
          container_node->SetContainerOpen (PR_FALSE);

          if (inner < child_count)
            break;
          if (best_access_count == 0)
            continue;

          rv = best_node->GetUri (url);
          if (NS_FAILED (rv))
            continue;
          rv = best_node->GetTitle (title);
          if (NS_FAILED (rv))
            continue;

          urls[n_found] = g_strdup (url.get ());
          titles[n_found] = g_strdup (title.get ());
          n_found ++;
        }

      g_signal_emit (self, history_signals[FAVORITES_RECEIVED_SIGNAL], 0,
                     urls, titles);
    }

  g_strfreev (titles);
  g_strfreev (urls);

  root_node->SetContainerOpen (PR_FALSE);

 rv_error:
  if (NS_FAILED (rv))
    {
      GError *error = NULL;
      mhs_error_set_from_nsresult (rv, &error);
      g_warning ("%s", error->message);
      g_error_free (error);
    }
  return FALSE;
}

static gboolean
mhs_history_get_favorites (MhsHistory *self,
                           GError **error)
{
  g_return_val_if_fail (MHS_IS_HISTORY (self), FALSE);

  MhsHistoryPrivate *priv = self->priv;

  /* Queue a new favorites query if we haven't already got one */
  if (priv->favorites_idle_handler == 0)
    priv->favorites_idle_handler
      = g_idle_add (mhs_history_favorites_cb, self);

  return TRUE;
}

static gboolean
mhs_history_remove_favorite (MhsHistory *self,
                             const gchar *uri_str,
                             GError **error)
{
  g_return_val_if_fail (MHS_IS_HISTORY (self), FALSE);

  nsresult rv;

  {
    nsCOMPtr<nsIURI> uri;
    rv = NS_NewURI (getter_AddRefs (uri), uri_str);
    if (NS_FAILED (rv))
      goto rv_error;

    nsCOMPtr<nsIBrowserHistory> browser_history
      = do_GetService (NS_GLOBALHISTORY2_CONTRACTID, &rv);
    if (NS_FAILED (rv))
      goto rv_error;

    nsCAutoString host;
    rv = uri->GetHost(host);
    if (NS_FAILED (rv))
      goto rv_error;

    // The favorites are grouped by host so to remove this favorite we
    // need to remove all history entries for the host of the URI
    rv = browser_history->RemovePagesFromHost (host, PR_FALSE);
    if (NS_FAILED (rv))
      goto rv_error;
  }

  return TRUE;

 rv_error:
  mhs_error_set_from_nsresult (rv, error);
  return FALSE;
}

static gboolean
mhs_history_clear_history (MhsHistory *self,
                           GError **error)
{
  g_return_val_if_fail (MHS_IS_HISTORY (self), FALSE);

  nsresult rv;

  nsCOMPtr<nsIBrowserHistory> browser_history
    = do_GetService (NS_GLOBALHISTORY2_CONTRACTID, &rv);

  if (NS_FAILED (rv))
    goto rv_error;

  rv = browser_history->RemoveAllPages ();
  if (NS_FAILED (rv))
    goto rv_error;

  return TRUE;

 rv_error:
  mhs_error_set_from_nsresult (rv, error);
  return FALSE;
}

static gboolean
mhs_history_get_is_page_pinned (MhsHistory *self,
                                const gchar *page_uri,
                                gboolean *is_page_pinned,
                                GError **error)
{
  g_return_val_if_fail (MHS_IS_HISTORY (self), FALSE);

  nsCOMPtr<nsINavBookmarksService> bookmark_service;
  nsCOMPtr<nsIURI> uri;
  PRBool is_page_pinned_pr;
  nsresult rv;

  bookmark_service = do_GetService (NS_NAVBOOKMARKSSERVICE_CONTRACTID, &rv);
  if (NS_FAILED (rv))
    goto rv_error;

  rv = NS_NewURI (getter_AddRefs (uri), page_uri);
  if (NS_FAILED (rv))
    goto rv_error;

  rv = bookmark_service->IsBookmarked (uri, &is_page_pinned_pr);
  if (NS_FAILED (rv))
    goto rv_error;

  *is_page_pinned = is_page_pinned_pr;

  return TRUE;

 rv_error:
  mhs_error_set_from_nsresult (rv, error);
  return FALSE;
}

static gboolean
mhs_history_pinned_pages_cb (gpointer data)
{
  nsresult rv = NS_OK;
  MhsHistory *self = (MhsHistory *) data;
  MhsHistoryPrivate *priv = self->priv;
  nsCOMPtr<nsINavBookmarksService> bookmark_service;
  nsCOMPtr<nsINavHistoryService> history_service;
  PRInt64 bookmarks_menu_folder;
  PRInt64 bookmark_id;
  PRInt32 index = 0;

  priv->pinned_pages_idle_handler = 0;

  bookmark_service = do_GetService (NS_NAVBOOKMARKSSERVICE_CONTRACTID, &rv);
  if (NS_FAILED (rv))
    goto rv_error;
  history_service = do_GetService (NS_NAVHISTORYSERVICE_CONTRACTID, &rv);
  if (NS_FAILED (rv))
    goto rv_error;

  rv = bookmark_service->GetBookmarksMenuFolder (&bookmarks_menu_folder);
  if (NS_FAILED (rv))
    goto rv_error;

  rv = bookmark_service->GetIdForItemAt (bookmarks_menu_folder, index++,
                                         &bookmark_id);
  if (NS_FAILED (rv))
    goto rv_error;

  while (bookmark_id != -1)
    {
      nsCOMPtr<nsIURI> uri;
      PRInt64 next_bookmark_id;

      rv = bookmark_service->GetIdForItemAt (bookmarks_menu_folder, index++,
                                             &next_bookmark_id);
      if (NS_FAILED (rv))
        goto rv_error;

      rv = mhs_history_emit_pinned_page_for_item (self,
                                                  bookmark_service,
                                                  history_service,
                                                  bookmark_id,
                                                  next_bookmark_id != -1);
      if (NS_FAILED (rv))
        goto rv_error;

      bookmark_id = next_bookmark_id;
    }

 rv_error:
  if (NS_FAILED (rv))
    {
      GError *error = NULL;
      mhs_error_set_from_nsresult (rv, &error);
      g_warning ("%s", error->message);
      g_error_free (error);
    }
  return FALSE;
}

static gboolean
mhs_history_get_pinned_pages (MhsHistory *self,
                              GError **error)
{
  g_return_val_if_fail (MHS_IS_HISTORY (self), FALSE);

  MhsHistoryPrivate *priv = self->priv;

  /* Queue a new pinned pages query if we haven't already got one */
  if (priv->pinned_pages_idle_handler == 0)
    priv->pinned_pages_idle_handler
      = g_idle_add (mhs_history_pinned_pages_cb, self);

  return TRUE;
}

static gboolean
mhs_history_pin_page (MhsHistory *self,
                      const gchar *uri_str,
                      const gchar *page_title,
                      GError **error)
{
  g_return_val_if_fail (MHS_IS_HISTORY (self), FALSE);

  nsCOMPtr<nsINavBookmarksService> bookmark_service;
  nsCOMPtr<nsIURI> uri;
  PRBool is_pinned;
  nsresult rv;

  bookmark_service = do_GetService (NS_NAVBOOKMARKSSERVICE_CONTRACTID, &rv);
  if (NS_FAILED (rv))
    goto rv_error;

  rv = NS_NewURI (getter_AddRefs (uri), uri_str);
  if (NS_FAILED (rv))
    goto rv_error;

  /* Don't do anything if the page is already pinned */
  rv = bookmark_service->IsBookmarked (uri, &is_pinned);
  if (NS_FAILED (rv))
    goto rv_error;

  if (!is_pinned)
    {
      PRInt64 bookmarks_menu_folder;
      PRInt64 bookmark_id;

      rv = bookmark_service->GetBookmarksMenuFolder (&bookmarks_menu_folder);
      if (NS_FAILED (rv))
        goto rv_error;

      rv = bookmark_service
        ->InsertBookmark (bookmarks_menu_folder,
                          uri, nsINavBookmarksService::DEFAULT_INDEX,
                          nsDependentCString (page_title),
                          &bookmark_id);
      if (NS_FAILED (rv))
        goto rv_error;
    }

  return TRUE;

 rv_error:
  mhs_error_set_from_nsresult (rv, error);
  return FALSE;
}

static gboolean
mhs_history_unpin_page (MhsHistory *self,
                        const gchar *uri_str,
                        GError **error)
{
  g_return_val_if_fail (MHS_IS_HISTORY (self), FALSE);

  nsCOMPtr<nsINavBookmarksService> bookmark_service;
  nsCOMPtr<nsIURI> uri;
  nsAutoTArray<PRInt64, 8> bookmark_ids;
  nsresult rv;
  unsigned int i;

  bookmark_service = do_GetService (NS_NAVBOOKMARKSSERVICE_CONTRACTID, &rv);
  if (NS_FAILED (rv))
    goto rv_error;

  rv = NS_NewURI (getter_AddRefs (uri), uri_str);
  if (NS_FAILED (rv))
    goto rv_error;

  rv = bookmark_service->GetBookmarkIdsForURITArray (uri, &bookmark_ids);
  if (NS_FAILED (rv))
    goto rv_error;

  for (i = 0; i < bookmark_ids.Length (); i++)
    {
      rv = bookmark_service->RemoveItem (bookmark_ids[i]);
      if (NS_FAILED (rv))
        goto rv_error;
    }

  return TRUE;

 rv_error:
  mhs_error_set_from_nsresult (rv, error);
  return FALSE;
}

static gboolean
mhs_history_unpin_all_pages (MhsHistory *self,
                             GError **error)
{
  g_return_val_if_fail (MHS_IS_HISTORY (self), FALSE);

  nsCOMPtr<nsINavBookmarksService> bookmark_service;
  PRInt64 bookmarks_menu_folder;
  nsresult rv;

  bookmark_service = do_GetService (NS_NAVBOOKMARKSSERVICE_CONTRACTID, &rv);
  if (NS_FAILED (rv))
    goto rv_error;

  rv = bookmark_service->GetBookmarksMenuFolder (&bookmarks_menu_folder);
  if (NS_FAILED (rv))
    goto rv_error;

  rv = bookmark_service->RemoveFolderChildren (bookmarks_menu_folder);
  if (NS_FAILED (rv))
    goto rv_error;

  return TRUE;

 rv_error:
  mhs_error_set_from_nsresult (rv, error);
  return FALSE;
}
